----------------------------------------------------------------------------
-- Haskell interface to Maple
--
-- (c) 2000
--     Hans-Wolfgang Loidl <hwloidl@cee.hw.ac.uk>
--     Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
--
-- $Id: maple2.hs,v 1.2 2000/11/08 16:00:41 schreine Exp $
----------------------------------------------------------------------------
{-# OPTIONS -#include "maple.h" #-}

module MapleAPI
  (MapleObject,                      -- Haskell representation of Maple things
   MapleCallable(..),
   nullMO, 
   mapleEval, mapleEvalN,            -- evaluating a Maple function
                                     -- conversion functions for basic types:
   string2MapleName, string2MapleExpr,
   int2MapleObject, mapleObject2Int,
   integer2MapleObject, mapleObject2Integer,
   char2MapleObject, mapleObject2Char,
   string2MapleObject, mapleObject2String,
                                     -- debugging only:
   , string2ByteArray, showMapleObject, printByteArray) where

-- @menu
-- * Imports::			
-- * Interface Specification::	
-- * MapleObject Type::		
-- * Classes and Instances::	
-- * Maple calls::		
-- * High level conversion routines::  
-- * Communication with Maple process::	
-- * Low level conversion routines::  
-- * Operations on ByteArrays::	
-- * Debugging routines::	
-- @end menu

-- @node Imports, Interface Specification
-- @section Imports

-- standard imports
-- import MyPackedString  -- same as PackedString, but with an NFData instance
import Strategies

-- prelude imports
import PrelAddr
import PrelBase
import PrelIOBase
import PrelPack ( new_ps_array,  freeze_ps_array,  write_ps_array )
import PrelConc
-- low level access to blocks of memory
import Addr	( readCharOffAddr, writeCharOffAddr )
import ByteArray
import MutableArray
import ST
import IO -- debugging  only
import Char

-- @node Interface Specification, MapleObject Type, Imports
-- @section Interface Specification

---------------------------------------------------------------------------
-- The interface consists of the following functions
--
-- * call a maple function "fun"
--
--    r  = mapleEval  "fun" [ a ]
--   [r] = mapleEvalN "fun" [ a ]
--
--    a: Int, Char, String, MapleObject
--    r: MapleObject
--
-- if Maple functions are called multiple times with same arguments,
-- it is more efficient to convert Int, Char, String to  MapleObject before
--
-- * conversion functions
--
--   m =    int2MapleObject i, i = mapleObject2Int    m
--   m =   char2MapleObject c, c = mapleObject2Char   m
--   m = string2MapleObject s, s = mapleObject2String m
--   m = string2MapleName s
--   m = string2MapleExpr s
--
--   m: MapleObject, I: i: Int, c: Char, s: String
---------------------------------------------------------------------------

-- @node MapleObject Type, Classes and Instances, Interface Specification
-- @section MapleObject Type

---------------------------------------------------------------------------
-- MapleObjects are things coming back from calls to Maple routines.
-- We use a PackedString as representation of Maple objects
-- This is a thin layer of abstraction over ByteArrays with handy
-- conversion functions from/to C-land
---------------------------------------------------------------------------
-- type MapleObject = PackedString
type MapleObject = ByteArray Int

-- @node Classes and Instances, Maple calls, MapleObject Type
-- @section Classes and Instances

-- class of things that can be passed to Maple
-- @cindex MapleCallable
class MapleCallable a where
  -- convert this Haskell object into a MapleObject
  toMapleObject :: a -> MapleObject
  -- and the other direction, too
  fromMapleObject :: MapleObject -> a
  -- write this Maple object to a file 
  writeToMaple :: Addr {-Handle-} -> a -> IO ()

-- a Haskell fixed precision integer
instance MapleCallable Int where
  toMapleObject = int2MapleObject
  fromMapleObject = mapleObject2Int
  writeToMaple fd i = writeStringToFile fd (show i)

-- a Haskell arbitrary precision integer
instance MapleCallable Integer where
  toMapleObject = integer2MapleObject
  fromMapleObject = mapleObject2Integer
  writeToMaple fd i = writeStringToFile fd (show i)

-- a Haskell character
instance MapleCallable Char where
  toMapleObject = char2MapleObject
  fromMapleObject = mapleObject2Char
  writeToMaple fd c = writeStringToFile fd (show c)

-- a Haskell string
instance MapleCallable String where
  toMapleObject = string2MapleObject
  fromMapleObject = mapleObject2String
  writeToMaple = writeStringToFile

-- a Maple object
instance MapleCallable MapleObject where
  toMapleObject = id
  fromMapleObject = id
  writeToMaple = writeByteArrayToFile

-- a MapleObject can be shown 
instance Show MapleObject where
  show = showMapleObject

instance NFData MapleObject where
  rnf (ByteArray _ _ ba#) = (C# (indexCharArray# ba# 0#))=='\0' `seq` ()

---------------------------------------------------------------------------
-- @node Constants
-- @section Constants
---------------------------------------------------------------------------

-- this MO means "undefined"
nullMO :: MapleObject
nullMO = string2ByteArray ""

---------------------------------------------------------------------------
-- @node Maple calls, High level conversion routines, Classes and Instances
-- @section Maple calls
---------------------------------------------------------------------------

-- ToDo:
--  . define a set of exceptions for possible errors on Maple side

-- This writes the fct name and the args (must have been converted to 
-- MapleObjects already) to the given handle in a format understood
-- by Maple (fd is actually a pipe to a Maple process)
-- @cindex mapleEvalCore
{-# INLINE mapleEvalCore #-}
mapleEvalCore :: Addr {-Handle-} -> String -> [MapleObject] -> IO ()
mapleEvalCore fd name args =
  do 
    writeStringToFile fd name
    mapM (\ arg -> (writeToMaple fd arg)) args
    return ()

-- Implements a Maple call; fct is a string; args anything MapleCallable
-- @cindex mapleEval
mapleEval :: (MapleCallable a) => String -> [a] -> MapleObject
mapleEval name args = 
  unsafePerformIO $
  do
    fd <- _ccall_ mapleOut :: IO Addr  -- grab the file descriptor from C-land
    mapleEvalM fd name args

-- pure, monadic version of the Maple call
{-# INLINE mapleEvalM #-}
mapleEvalM :: (MapleCallable a) => Addr -> String -> [a] -> IO MapleObject
mapleEvalM fd name args =                 -- <maple fct> <list of maple args>
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleEvalM"
    hFlush stderr
#endif
    let 
      n = length args                  -- number of args to maple fct call
      mos = map toMapleObject args     -- convert all args to MapleObjects
    mylock                    `demanding` seqList rwhnf mos
    (_ccall_ mapleStartEval n) 
    mapleEvalCore fd name mos          -- write fct and arg names
    _ccall_ mapleEndEval               -- write call-epilogue
    mapleEvalWait                      -- wait for result
    mr <- _ccall_ mapleResult :: IO Addr -- grab result
    let mo = mapleExtractMO_ret mr
    myunlock `demanding` rnf mo
    return mo                          -- and pull into Haskell-land

-- Implements a Maple call; returns a list of results
-- @cindex mapleEvalN
mapleEvalN :: (MapleCallable a) => String -> [a] -> [MapleObject]
mapleEvalN name args = 
  unsafePerformIO $ 
  do 
    fd <- _ccall_ mapleOut :: IO Addr  -- grab the file descriptor from C-land
    mapleEvalNM fd name args

-- pure, monadic version of the Maple call
{-# INLINE mapleEvalNM #-}
mapleEvalNM :: (MapleCallable a) => Addr -> String -> [a] -> IO [MapleObject]
mapleEvalNM fd name args =
  do 
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleEvalNM"
    hFlush stderr
#endif
    let 
      n = length args                  -- number of args to maple fct call
      mos = map toMapleObject args     -- convert all args to MapleObjects
    mylock                    `demanding`    seqList rwhnf mos -- force output
    (_ccall_ mapleStartEvalN n) 
    mapleEvalCore fd name mos          -- write fct and arg names
    _ccall_ mapleEndEval               -- write call-epilogue
    mapleEvalWait                      -- wait for result
    m <- _ccall_ mapleResultN :: IO Int  -- how many results?
    l <- mapM ( \ _ -> do
                  mr <- _ccall_ mapleNextResult :: IO Addr  -- grab next result
                  let x = mapleExtractMO_ret mr
                  return $! x) [1..m]  -- must be strict to eval every
    myunlock `demanding` rnf l
    return(l)                             --  list elem!

-- wait for Haskell resultant to be available, non-blocking for system
mapleEvalWait :: IO ()
mapleEvalWait =
  do
    b <- _ccall_ mapleResultAvailable :: IO Int
    if b == 0 then yield {- appendFile "bonzo" "rubbish" -} >> mapleEvalWait else return ()

---------------------------------------------------------------------------
-- @node High level conversion routines, Communication with Maple process, Maple calls
-- @section High level conversion routines
---------------------------------------------------------------------------

-- a -> MapleObject

-- convert a Haskell Int to a MapleObject
-- @cindex int2MapleObject
int2MapleObject :: Int -> MapleObject
int2MapleObject = unsafePerformIO . int2MapleObjectM

int2MapleObjectM :: Int -> IO MapleObject
int2MapleObjectM n =
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> int2MapleObjectM"
    hFlush stderr
#endif
    mylock `demanding` rnf n
    addr <- (_ccall_ mapleInt n) :: IO Addr
    let mo = mapleExtractMO_ret addr
    myunlock `demanding` rnf mo
    return mo

-- convert a Haskell arbitrary precision integer to a MapleObject
integer2MapleObject :: Integer -> MapleObject
integer2MapleObject = unsafePerformIO . integer2MapleObjectM

integer2MapleObjectM :: Integer -> IO MapleObject
integer2MapleObjectM n =
  do 
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> integer2MapleObjectM"
    hFlush stderr
#endif
    mylock `demanding` rnf n
    let
      ba@(ByteArray _ _ ba#) = string2ByteArray $ show n
      x = lengthMO ba
    addr <- (_ccall_ mapleInteger ba# x) :: IO Addr
    -- _ccall_ free cs
    let mo = mapleExtractMO_ret addr
    myunlock `demanding` rnf mo
    return mo

-- convert a MapleObject to a Haskell arbitrary precision Integer
-- UNTESTED CODE
mapleObject2Integer :: MapleObject -> Integer
mapleObject2Integer = unsafePerformIO . mapleObject2IntegerM

mapleObject2IntegerM :: MapleObject -> IO Integer
mapleObject2IntegerM mo@(ByteArray _ _ ba#) = 
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleObject2IntegerM"
    hFlush stderr
#endif
    mylock `demanding` rnf mo
    let 
      addr = mo2CString mo
      x = lengthMO mo
    ret <- (_ccall_ haskellString addr x) :: IO Addr
    _ccall_ free addr
    let i = read (byteArray2String (mapleExtractMO_ret ret)) :: Integer
    myunlock `demanding` rnf i
    return i

-- convert a MapleObject to a Haskell Int
-- @cindex mapleObject2Int
mapleObject2Int :: MapleObject -> Int
mapleObject2Int = unsafePerformIO . mapleObject2IntM

mapleObject2IntM :: MapleObject -> IO Int
mapleObject2IntM mo@(ByteArray _ _ mo#) = 
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleObject2Int"
    hFlush stderr
#endif
    mylock `demanding` rnf mo
    let 
      n = lengthMO# mo#
    ret <- (_ccall_ haskellInt mo# n) :: IO Addr
    -- _ccall_ free addr
    let i = mapleExtractInt_ret ret :: Int
    myunlock `demanding` rnf i
    return i

-- convert a Haskell Char to a MapleObject
-- @cindex char2MapleObject
char2MapleObject :: Char -> MapleObject
char2MapleObject = unsafePerformIO . char2MapleObjectM

char2MapleObjectM :: Char -> IO MapleObject
char2MapleObjectM n =
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> char2MapleObjectM"
    hFlush stderr
#endif
    mylock `demanding` rnf n
    addr <- (_ccall_ mapleChar n) :: IO Addr
    let mo = mapleExtractMO_ret addr
    myunlock `demanding` rnf mo
    return mo

-- convert a MapleObject to a Haskell Char
-- @cindex mapleObject2Char
mapleObject2Char :: MapleObject -> Char
mapleObject2Char = unsafePerformIO . mapleObject2CharM

mapleObject2CharM :: MapleObject -> IO Char
mapleObject2CharM mo@(ByteArray _ _ mo#) = 
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleObject2CharM"
    hFlush stderr
#endif
    mylock `demanding` rnf mo
    let 
      -- addr = psToCString mo
      n = lengthMO# mo#
    ret <- (_ccall_ haskellChar mo# n) :: IO Addr
    -- _ccall_ free addr
    let ch = mapleExtractChar_ret ret
    myunlock `demanding` rnf ch
    return ch

-- convert a MapleObject to a Haskell String
-- @cindex mapleObject2String
mapleObject2String :: MapleObject -> String
mapleObject2String = unsafePerformIO . mapleObject2StringM

mapleObject2StringM :: MapleObject -> IO String
mapleObject2StringM mo@(ByteArray _ _ ba#) = 
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> mapleObject2StringM"
    hFlush stderr
#endif

    mylock `demanding` rnf mo
    let 
      addr = mo2CString mo
      x = lengthMO mo
    ret <- (_ccall_ haskellString addr x) :: IO Addr
    _ccall_ free addr
    let str = byteArray2String (mapleExtractMO_ret ret)
    myunlock `demanding` rnf str
    return str

-- convert a MapleObject to a Haskell String
-- @cindex string2MapleExpr
string2MapleExpr :: String -> MapleObject
string2MapleExpr = byteArray2MapleExpr . string2ByteArray
{-
  unsafePerformIO $
  do
    putStr ("string2MapleExpr of |" ++ str ++ "| yields bytearray |")
    let mo = byteArray2MapleExpr $ string2ByteArray $ str
    -- printByteArray mo
    -- hPutStrLn stderr "|"
    return mo
-}
    
{-
  unsafePerformIO $
  do 
    mo <- stToIO (string2ByteArrayST str)
    let 
      cstr@(A# cstr#) = mo2CString mo
      (I# x#) = strlen# cstr#
    addr <- (_ccall_ mapleExpr cstr# x#) :: IO Addr
    return (mapleExtractMO_ret addr)
-}

-- convert a Haskell packed string to a MapleObject
byteArray2MapleExpr :: ByteArray Int -> MapleObject
byteArray2MapleExpr = unsafePerformIO . byteArray2MapleExprM

{-# INLINE byteArray2MapleExprM #-}
byteArray2MapleExprM :: ByteArray Int -> IO MapleObject
byteArray2MapleExprM mo@(ByteArray _ _ ba#) =
  do 
#if defined(DEBUG_LOCK)
    hPutStrLn stderr "-> byteArray2MapleExprM"
    hFlush stderr
#endif
    mylock `demanding` rnf mo
    let
      x = lengthMO# ba#
    addr <- (_ccall_ mapleExpr ba# x) :: IO Addr
    let mo = mapleExtractMO_ret addr
    myunlock `demanding` rnf mo
    return mo

-- convert a MapleObject to a Haskell String
string2MapleObject x = string2MapleExpr ("\"" ++ x ++ "\"")

-- convert a MapleObject to a Haskell String
string2MapleName x = string2MapleExpr ("`" ++ x ++ "`")

---------------------------------------------------------------------------
-- @node Communication with Maple process, Low level conversion routines, High level conversion routines
-- @section Communication with Maple process
--
-- Writing a MapleObject directly to the pipe connecting with the Maple proc
---------------------------------------------------------------------------

-- @cindex writeStringToFile
writeStringToFile :: Addr {-Handle-} -> String -> IO ()
writeStringToFile fd str =
  do
#if defined(DEBUG_WRITE)
    hPutStrLn stderr str
    hFlush stdout
#endif
    let ba = string2ByteArray str
#if defined(DEBUG_WRITE)
    printByteArray ba                -- debugging
#endif
    writeByteArrayToFile fd ba

{-
writePackedStringToFile :: Addr -> PackedString -> IO ()
writePackedStringToFile fd str = do
  let 
    addr = psToCString str
    x = lengthPS str
  _ccall_ writeLine fd addr x
  _ccall_ free addr
-}

-- @cindex writeByteArrayToFile
writeByteArrayToFile :: Addr {-Handle-} -> ByteArray Int -> IO ()
writeByteArrayToFile fd mo@(ByteArray _ _ ba#) = do
  let 
    addr = mo2CString mo
    x = lengthMO mo
    -- ASSERT: length str == lengthMO mo
#if defined(DEBUG_MO_SANITY)
  foo <- if (not (isSaneMO mo))
           then ioError (userError ("writeByteArrayToFile: invalid maple bytestream: |" ++ (byteArray2String mo) ++ "|"))
           else return mo
#endif
  _ccall_ writeLine fd addr x
  _ccall_ free addr

---------------------------------------------------------------------------
-- @node Low level conversion routines, Operations on ByteArrays, Communication with Maple process
-- @section Low level conversion routines
---------------------------------------------------------------------------

-- get an Int value out of a MapleRet structure
-- @cindex mapleExtractInt_ret
mapleExtractInt_ret :: Addr -> Int
mapleExtractInt_ret tup_addr = 
  let 
    (A# tup_addr#) = tup_addr
    ret#  = indexIntOffAddr# tup_addr# 0#  :: Int#  -- return code 
  in
  if (ret# ==# 0#)
    then mapleExtractInt (A# (int2Addr# ((addr2Int# tup_addr#) +# 4#)))
    else error "error in mapleExtractInt_ret"

-- convert a Haskell-side Maple object into a C-allocated Maple object
-- @cindex mapleExtractInt
mapleExtractInt :: Addr -> Int
mapleExtractInt (A# addr#) = I# (indexIntOffAddr# addr# 0#)

-- get an Char value out of a MapleRet structure
-- @cindex mapleExtractChar_ret
mapleExtractChar_ret :: Addr -> Char
mapleExtractChar_ret tup_addr = 
  let 
    (A# tup_addr#) = tup_addr
    ret#  = indexIntOffAddr# tup_addr# 0#  :: Int#  -- return code 
  in
  if (ret# ==# 0#)
    then mapleExtractChar (A# (int2Addr# ((addr2Int# tup_addr#) +# 4#)))
    else error "error in mapleExtractChar_ret"

-- convert a Haskell-side Maple object into a C-allocated Maple object
-- @cindex mapleExtractChar
mapleExtractChar :: Addr -> Char
mapleExtractChar (A# addr#) = C# (indexCharOffAddr# addr# 0#)

-- convert a Haskell-side Maple object into a C-allocated Maple object
-- @cindex mapleExtractString_ret
mapleExtractString_ret :: Addr -> String
mapleExtractString_ret = byteArray2String . mapleExtractMO_ret

-- convert a result from a maple call as defined by MapleRet strcut, i.e. with
-- leading return code, into a Haskell-allocated packed string
-- @cindex mapleExtractMO_ret
mapleExtractMO_ret :: Addr -> MapleObject
mapleExtractMO_ret tup_addr = 
  let 
    (A# tup_addr#) = tup_addr
    ret#  = indexIntOffAddr# tup_addr# 0#  :: Int#  -- return code 
  in
    if (ret# ==# 0#)
      then mapleExtractMO (A# (int2Addr# ((addr2Int# tup_addr#) +# 4#)))
      else error "error in mapleExtractMO_ret"
    
-- convert a C-side Maple object into a Haskell-allocated Maple object
-- @cindex mapleExtractMO
mapleExtractMO :: Addr -> MapleObject
mapleExtractMO tup_addr = 
  unsafePerformIO $
  do
  let 
    -- mapleExtractPS elems of C struct:
    (A# tup_addr#) = tup_addr
    n#    = indexIntOffAddr# tup_addr# 0#      :: Int#  -- size
    -- HACK: replace with proper address arithmetic indep of word size
    str_addr# = int2Addr# ((addr2Int# tup_addr#) +# 4#)
    -- tup_addr `plusAddr` (AddrOff (I# 1#)) 
    addr#    = indexAddrOffAddr# str_addr# 0#  :: Addr#  -- pointer to data
#if defined(DEBUG_MO_SANITY)
  mo <- stToIO (cBytes2ByteArrayST (I# n#) (A# addr#))
  if (not (isSaneMO mo))
    then ioError (userError ("mapleExtractMO: invalid maple bytestream returned: |" ++ (byteArray2String mo) ++ "|"))
    else return mo
#else
  stToIO (cBytes2ByteArrayST (I# n#) (A# addr#))	
#endif

#if defined(DEBUG_MO_SANITY)
isSaneMO :: MapleObject -> Bool
isSaneMO mo = let str = byteArray2String mo
              in  
	        and (map ( \ c -> isAscii c && (not (isControl c))) str)
#endif

---------------------------------------------------------------------------
-- @node Operations on ByteArrays, Debugging routines, Low level conversion routines
-- @section Operations on ByteArrays
--
-- Basic operations on the representation of MapleObjects as ByteArrays
-- Most of these routines have been (aehm) borrowed from the PackedString
-- module. However, we use a simpler type, just a ByteArray.
---------------------------------------------------------------------------

-- length of the ByteArray
-- @cindex lengthMO
lengthMO   :: MapleObject -> Int
lengthMO mo@(ByteArray _ _ mo#) = I# (lengthMO# mo#)

{-# INLINE lengthMO# #-}
-- @cindex lengthMO#
lengthMO# :: ByteArray# -> Int#
lengthMO# mo = findNull# 0# mo

{-# INLINE findNull# #-}
-- @cindex findNull#
findNull# :: Int# -> ByteArray# -> Int#
findNull# i# frozen#
     | ch# `eqChar#` '\0'# = i# -- everything upto the sentinel
     | otherwise	   = findNull# (i# +# 1#) frozen#
    where
     ch#  = indexCharArray# frozen# i#

{-# INLINE strlen# #-}
-- @cindex strlen#
strlen# :: Addr# -> Int
strlen# a
  = unsafePerformIO $
    _ccall_ strlen (A# a)  >>= \ len ->
    return len

-----------------------

-- @cindex indexMO
indexMO :: MapleObject -> Int -> Char
indexMO mo@(ByteArray _ _ mo#) (I# n#) = C# (indexMO# mo# n#)

{-# INLINE indexMO# #-}
-- @cindex indexMO#
indexMO# :: ByteArray# -> Int# -> Char#
indexMO# mo n#
  = --ASSERT (n >=# 0# && n <# i)	-- error checking: my eye!  (WDP 94/10)
    indexCharArray# mo n#

-- Packing/Unpacking

-- write the contents of the MapleObject into a malloc-ed C structure
-- @cindex mo2CString
mo2CString :: MapleObject -> Addr
mo2CString mo@(ByteArray _ _ mo#) = 
  unsafePerformIO $
  do
    let l@(I# l#) = lengthMO mo
    stuff <- _ccall_ malloc ((l+1) * (``sizeof(char)''))
    let
     fill_in n# i#
      | n# ==# 0# = do
         writeCharOffAddr stuff (I# i#) (C# (chr# 0#))
         return ()
      | otherwise = do
         let ch#  = indexCharArray# mo# i#
         writeCharOffAddr stuff (I# i#) (C# ch#)
         fill_in (n# -# 1#) (i# +# 1#)
    fill_in l# 0#
    return stuff    

-- shouldn't be necessary!
-- ToDo: merge string2ByteArray and cBytes2ByteArray (same structure, different arg)

-- @cindex string2ByteArray
string2ByteArray :: [Char] -> ByteArray Int
string2ByteArray str = unsafePerformIO (string2ByteArrayIO str)
{-
  unsafePerformIO $
  do
   putStrLn ("Doing a string2ByteArray on string |" ++ str ++ "|")
   hFlush stdout
   let ba = runST (string2ByteArrayST str)
   printByteArray ba 
   return ba
-}

string2ByteArrayIO :: [Char] -> IO (ByteArray Int)
string2ByteArrayIO str =
 do 
  let len@(I# length#) = length str
   {- 
    allocate an array that will hold the string
    (not forgetting the NUL byte at the end)
   -}
  -- putStrLn ("string2ByteArrayIO of " ++ (show str) ++ " length " ++ (show len))
  ch_array <- stToIO (new_ps_array (length# +# 1#))
  -- fill in the contents of the bytearray from "str"
  stToIO (fill_in ch_array 0# str)
  -- freeze the puppy:
  ba@(ByteArray _ _ frozen#) <- stToIO (freeze_ps_array ch_array (length# +# 1#))
  -- return just the unboxed bytearray
  -- putStrLn ("Done printing that bloody string.")
  return ba
  where
   fill_in :: MutableByteArray s Int -> Int# -> [Char] -> ST s ()
   fill_in arr_in# idx [] =
    do 
      write_ps_array arr_in# idx (chr# 0#)
      return ()
 
   fill_in arr_in# idx (C# c : cs) =
    do 
      -- putChar (C# c) 
      write_ps_array arr_in# idx c
      fill_in arr_in# (idx +# 1#) cs

-- C mem-block -> Haskell ByteArray

cBytes2ByteArrayST :: Int -> Addr -> ST s (ByteArray Int)
cBytes2ByteArrayST (I# length#) (A# addr) =
 do
  {- 
    allocate an array that will hold the string
    (not forgetting the NUL byte at the end)
  -}
  ch_array <- new_ps_array (length# +# 1#)
  fill_in ch_array 0#
  ba <- freeze_ps_array ch_array (length# +# 1#)
  return ba
  where
    fill_in :: MutableByteArray s Int -> Int# -> ST s ()
    fill_in arr_in# idx
      | idx ==# length#
      = write_ps_array arr_in# idx (chr# 0#) >>
	return ()
      | otherwise
      = case (indexCharOffAddr# addr idx) of { ch ->
	write_ps_array arr_in# idx ch >>
	fill_in arr_in# (idx +# 1#) }

-- builds a String out of a MapleObject

-- @cindex byteArray2String
byteArray2String :: ByteArray Int -> [Char]
byteArray2String ba@(ByteArray _ _ ba#) = unpack 0#
 where
    len# = findNull# 0# ba#
    unpack nh
      | nh >=# len#  = []
      | otherwise   = C# ch# : unpack (nh +# 1#)
      where
	ch# = indexCharArray# ba# nh

---------------------------------------------------------------------------
-- lock/unlock operations
-- THIS IS JUST A BAD HACK; SHOULD BE DONE VIA MVARS
---------------------------------------------------------------------------

mylock :: IO ()
mylock = 
  do 
    l <- _casm_ ``{ extern int lockVar; %r = lockVar; lockVar = 1; }'' :: IO Int
#if defined(DEBUG_LOCK)
    hPutStrLn stderr ("LOCK0; found value " ++ (show l) )
    hFlush stdout
#endif
    mylock' l

mylock' :: Int -> IO ()
mylock' l = 
  do
    if (l==1)
      then do
             yield 
             l' <- _casm_ ``{ extern int lockVar; %r = lockVar; lockVar = 1; }''  :: IO Int
#if defined(DEBUG_LOCK)
             hPutStrLn stderr ("LOCK1; found value " ++ (show l) )
             hFlush stdout
#endif
             mylock' l'
      else return ()

myunlock :: IO ()
myunlock = 
  do
#if defined(DEBUG_LOCK)
    hPutStrLn stderr ("UNLOCK")
    hFlush stdout
#endif
    _casm_ ``{ extern int lockVar; lockVar = 0; }'' :: IO ()

---------------------------------------------------------------------------
-- @node Debugging routines,  , Operations on ByteArrays
-- @section Debugging routines
---------------------------------------------------------------------------

showMapleObject :: MapleObject -> [Char]
showMapleObject mo@(ByteArray _ _ ba#) = "|" ++ byteArray2String mo ++ "|"
  -- "DBG> Size: " ++  (show (lengthMO mo)) ++ "; Contents: " ++ (byteArray2String mo)

printByteArray :: ByteArray Int -> IO ()
printByteArray (ByteArray _ _ ba#) = 
 do
   putStrLn "Printing a ByteArray (enclosed in '|'s):"
   putChar '|'
   print'em 0#
   putStrLn "|"
   hFlush stdout
   where
    len# = lengthMO# ba#
    print'em nh
      | nh >=# len# = return ()
      | otherwise   = putChar (C# ch#) >> print'em (nh +# 1#)
      where
	ch# = indexCharArray# ba# nh

----------------------------------------------------------------------------
-- $Id: maple2.hs,v 1.2 2000/11/08 16:00:41 schreine Exp $
----------------------------------------------------------------------------
